/* ----------------------------------------------------------------------
 * STRONGHOLD WEB SERVER
 * C2Net Europe Ltd 1999,2000
 *
 * Hooks and modificiations to mod_ssl to allow backward compatibility 
 * with old Stronghold 2 directives and features
 * ----------------------------------------------------------------------
 */

#include "mod_ssl.h"

static int ssl_sh2_configure_server_hook(server_rec *s, pool *p, SSLSrvConfigRec *sc);
static int ssl_sh2_update_child_status(short_score *ss, request_rec *r);
static int ssl_sh2_get_scoreboard_string(short_score *score_record, request_rec *r);

#define PRODUCT_SH3

#include "util_date.h"
#ifdef USE_HARDWARE 
#include <openssl/hardware.h>
#endif

static int check_exec(char *program)
{
    struct stat st;

    if(stat(program, &st) == -1) {
        return(0);
    }
    if(S_ISREG(st.st_mode) && (st.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))) {
        return(1);
    }
    else {
	return(0);
    }
}


/* For mod_ssl 2.3.0 we need to define SSL_VENDOR then we can use the hooks below
 * also need to define SSL_VENDOR_OBJS
 */

void ssl_vendor_register(void) 
{
  ap_hook_register("ap::mod_ssl::vendor::update_child_status",ssl_sh2_update_child_status,AP_HOOK_NOCTX);
  ap_hook_register("ap::mod_ssl::vendor::get_scoreboard_string",ssl_sh2_get_scoreboard_string,AP_HOOK_NOCTX);
  ap_hook_register("ap::mod_ssl::vendor::configure_server",ssl_sh2_configure_server_hook,AP_HOOK_NOCTX);
}

void ssl_vendor_unregister(void) 
{
  ap_hook_unregister("ap::mod_ssl::vendor::update_child_status",ssl_sh2_update_child_status);
  ap_hook_unregister("ap::mod_ssl::vendor::get_scoreboard_string",ssl_sh2_get_scoreboard_string);
  ap_hook_unregister("ap::mod_ssl::vendor::configure_server",ssl_sh2_configure_server_hook);
}

static int ssl_sh2_update_child_status(short_score *ss, request_rec *r)
{
    SSLSrvConfigRec *sc = NULL;
    SSL *ssl;

    if (!r || !r->server) 
	return DECLINED;
    if ((sc = mySrvConfig(r->server)) == NULL)
	return DECLINED;

    /* ssl could be NULL if we're doing http to a SSL vhost for example */

    if ((!sc->bEnabled) ||
	((ssl = ap_ctx_get(r->connection->client->ctx, "ssl")) == NULL)) {
	ss->ssl_version = 0;
	ss->ssl_cipher = 0;
	return OK;
    }
    if (ssl->session && ssl->session->cipher)
	ss->ssl_cipher = ssl->session->cipher->id;
    else
	ss->ssl_cipher = 0;
    ss->ssl_version = ssl->version;
    return OK;
}

/* we want to display the SSL version and cipher, but we only have internal OpenSSL
 * identifiers for both.  So we have to stuff them into the right structures and
 * do some calls in order to work out what they meant.  This saves us having to
 * put strings into the scoreboard making it bigger
 */

static int ssl_sh2_get_scoreboard_string(short_score *score_record, request_rec *r)
{
    SSL s;
    SSL_METHOD *meth;
    SSL_CIPHER cipher_in;
    SSL_CIPHER *cipher_out;
    char p[4];

    if (!score_record->ssl_version) {
	ap_rputs("N",r);
	return OK;
    }
    s.version = score_record->ssl_version;
    ap_rputs(SSL_get_version(&s),r);

    switch(s.version) {
    case SSL2_VERSION: meth = SSLv2_method();  break;
    case SSL3_VERSION: meth = SSLv3_method();  break;
    case TLS1_VERSION: meth = TLSv1_method();  break;
    default: return DECLINED;
    }
    cipher_in.id = score_record->ssl_cipher;
    meth->put_cipher_by_char(&cipher_in,p);
    cipher_out = meth->get_cipher_by_char(p);     
    if (cipher_out && cipher_out->name) 
	ap_rprintf(r," %s", cipher_out->name);
    return OK;
}

/***********************************************************************
 * We want to mail a warning if your certificate is about to expire
 *
 * Used to be in Stronghold 2.4.2
 * MJC 26.4.00
 ***********************************************************************/

static time_t ASN1_UTCTIME_time_t(ASN1_UTCTIME *utc)
{
    struct tm tm;
    int i = utc->length;

    if(i < 10)
	return -1;
    for(i = 0; i < 10; i++)
	if((utc->data[i] > '9') || (utc->data[i] < '0'))
	    return -1;

    tm.tm_year = (utc->data[0]-'0') * 10 + (utc->data[1]-'0');

    /* Deal with Year 2000 like eay did */
    if(tm.tm_year < 70)
	tm.tm_year += 100;

    tm.tm_mon = (utc->data[2]-'0') * 10 + (utc->data[3]-'0') - 1;
    tm.tm_mday = (utc->data[4]-'0') * 10 + (utc->data[5]-'0');
    tm.tm_hour = (utc->data[6]-'0') * 10 + (utc->data[7]-'0');
    tm.tm_min = (utc->data[8]-'0') * 10 + (utc->data[9]-'0');
    tm.tm_sec = (utc->data[10]-'0') * 10 + (utc->data[11]-'0');
    tm.tm_isdst = -1;

    return ap_tm2sec(&tm);
}

static void mail_warning(server_rec *s, int days)
{
    char *sm_path;
    FILE *sm;
    char buf[MAX_STRING_LEN];
    int pid, status;

    /* Check /usr/lib/sendmail, and then /usr/sbin/sendmail */
    sm_path = "/usr/lib/sendmail";
    
    if(!check_exec(sm_path)) {
	sm_path = "/usr/sbin/sendmail";
	if(!check_exec(sm_path))
	    return;
    }

    if(!(pid = fork())) {
	fclose(stderr);
	fclose(stdout);
	sprintf(buf, "%s -oi -oem -t", sm_path);

	sm = popen(buf, "w");
	if (days > 0) {
	    fprintf(sm, "To: %s\n"
		    "Subject: WARNING: The certificate for %s expires in %d day(s)\n"
		    "\n",
		    s->server_admin,
		    s->server_hostname ? s->server_hostname : "(null)",
		    days);
	} else {
	    fprintf(sm, "To: %s\n"
		    "Subject: The certificate for %s has expired\n"
		    "\n",
		    s->server_admin,
		    s->server_hostname ? s->server_hostname : "(null)");
	}

	/* Mail the message */
	fprintf(sm,
		"\n"
		"----------------- WARNING -----------------\n"
		"\n"
		"Your certificate needs to be renewed. Use the\n"
		"renewal program supplied with Stronghold to\n"
		"generate a renewal CSR for submission to your\n"
		"certificate authority.\n"
		"\n"
		"If the certificate expires browsers will not\n"
		"be able to connect to your site using SSL.\n"
		"\n"
		"----------------- WARNING -----------------\n");
	pclose(sm);
	exit(0);
    }
    waitpid(pid, &status, 0);
}
  
static int ssl_sh2_check_rsa_cert(server_rec *s, X509 *cert)
{
    ASN1_UTCTIME *notAfter, *notBefore;
    time_t begin, end;
    time_t cur_time;
    char *buf1, *buf2;
    int days;
	  
    /* Spew the expiration date here */
    notAfter = X509_get_notAfter(cert);
    notBefore = X509_get_notBefore(cert);

    /* Now actually check to see if it is near expiration */
    begin = ASN1_UTCTIME_time_t(notBefore);
    if (begin == -1) 
	return DECLINED;
    end = ASN1_UTCTIME_time_t(notAfter);
    if (end == -1) 
	return DECLINED;
    
    printf( "%s:%d: ",s->server_hostname?s->server_hostname: "(null)",s->port);

    buf1 = ctime(&begin);
    buf1[strlen(buf1)-1] = '\0';
    printf( "%s - ",buf1);

    buf2 = ctime(&end);
    buf2[strlen(buf2)-1] = '\0';

    printf("%s\n", buf2);

    cur_time = time(NULL);
    if(cur_time == -1)
	return DECLINED;

    /* Check */
    if(begin > cur_time) {
	printf("%s:%d: WARNING: Certificate not yet valid\n",
	       s->server_hostname ? s->server_hostname : "(null)",s->port);
    }
    if(end < cur_time) {
	printf("%s:%d: WARNING: Certificate EXPIRED\n",
	       s->server_hostname ? s->server_hostname : "(null)",s->port);
	mail_warning(s, -1);
    }
    days = (end - cur_time) / (3600 * 24);
    if(days < 30) {
	printf("%s:%d: WARNING: Certificate expires in %d day(s)\n",
	       s->server_hostname ? s->server_hostname : "(null)", s->port,
	       days);
	/* Send a warning if days is 30, 14, or less than 7 */
	if(days == 30 || days == 14 || days < 7)
	    mail_warning(s, days);
    }
    return OK;
}

static int ssl_sh2_configure_server_hook(server_rec *s, pool *p, SSLSrvConfigRec *sc)
{    
    if (sc->pPublicCert[SSL_AIDX_RSA])
	ssl_sh2_check_rsa_cert(s, sc->pPublicCert[SSL_AIDX_RSA]);
    return OK;
}
